/**
 * Copyright (C) 2022 Xilinx, Inc
 * Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

// ------ I N C L U D E   F I L E S -------------------------------------------
// Local - Include Files
#include <Validate.h> // from library

#include "SubCmdValidate.h"
#include "../common/MA35Device.h"
#include "../common/MAHelpMenus.h"
#include "../common/MAUtilitiesCore.h"
#include "../common/MAUtilities.h"
namespace XBU = XBUtilities;

// 3rd Party Library - Include Files
#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>
#include <boost/format.hpp>
#include <boost/program_options.hpp>
#include <boost/property_tree/json_parser.hpp>
namespace po = boost::program_options;

// System - Include Files
#include <iostream>

// =============================================================================

// ------ L O C A L   F U N C T I O N S ---------------------------------------

// ----- C L A S S   M E T H O D S -------------------------------------------

SubCmdValidate::SubCmdValidate(bool _isHidden, bool _isDepricated, bool _isPreliminary) : SubCmd("validate", "Validates the basic MA35 functionality") {
  const std::string longDescription = "Validates the given device by executing the platform's validate executable.";
  setLongDescription(longDescription);
  setExampleSyntax("");
  setIsHidden(_isHidden);
  setIsDeprecated(_isDepricated);
  setIsPreliminary(_isPreliminary);
}

void SubCmdValidate::execute(const SubCmdOptions& _options) const

{
  // -- Build up the format options
  const std::string formatOptionValues = XBU::create_suboption_list_string(Report::getSchemaDescriptionVector());

  // -- Retrieve and parse the subcommand options -----------------------------
  std::vector<std::string> device;
  std::string              sFormat;
  std::string              sOutput;
  bool                     help = false;

  po::options_description commonOptions("Commmon Options");
  commonOptions.add_options()("device,d", boost::program_options::value<decltype(device)>(&device)->multitoken(),
      "The device of interest. This is specified as follows:\n"
      "  <BDF> - Bus:Device.Function (e.g., 0000:d8:00.0)")("format,f", boost::program_options::value<decltype(sFormat)>(&sFormat),
      (std::string("Report output format. Valid values are:\n") + formatOptionValues).c_str())("output,o", boost::program_options::value<decltype(sOutput)>(&sOutput),
      "Direct the output to the given file") ("help", boost::program_options::bool_switch(&help), "Help to use this sub-command");

  po::options_description hiddenOptions("Hidden Options");

  // Parse sub-command ...
  po::variables_map vm;
  process_arguments(vm, _options, commonOptions, hiddenOptions);

  // Check to see if help was requested or no command was found
  if (help) {
    printHelp(commonOptions, hiddenOptions);
    return;
  }

  // -- Process the options --------------------------------------------
  Report::SchemaVersion    schemaVersion = Report::SchemaVersion::unknown; // Output schema version
  std::vector<std::string> param;
  // When  is specified, make sure an accompanying output file is also specified
  if (!sFormat.empty() && sOutput.empty()) {
    std::cerr << "ERROR: Please specify an output file to redirect the json to" << std::endl;
    throw std::runtime_error("Operation Canceled");
  }

  if (sFormat.empty())
    sFormat = "json";

  // Output Format
  schemaVersion = Report::getSchemaDescription(sFormat).schemaVersion;
  if (schemaVersion == Report::SchemaVersion::unknown)
    throw std::runtime_error((boost::format("Unknown output format: '%s'") % sFormat).str());

  // Output file
  if (!sOutput.empty() && !XBU::getForce() && boost::filesystem::exists(sOutput))
    throw std::runtime_error((boost::format("Output file already exists: '%s'") % sOutput).str());

  // Collect all of the devices of interest
  std::set<std::string> deviceNames;
  DeviceCollection      devices;
  for (const auto& deviceName : device)
    deviceNames.insert(boost::algorithm::to_lower_copy(deviceName));

  XBU::collect_devices(deviceNames, true /*inUserDomain*/, devices);

  // enforce 1 device specification
  if (devices.empty() || devices.size() > 1) {
    std::cerr << "\nERROR: Please specify a single device using --device option\n\n";
    XBU::list_devices(std::cerr, true);
    throw std::runtime_error("Operation Cancelled");
  }

  boost::property_tree::ptree pt;
  XBUtilities::get_device_info(devices[0], pt);
  if (pt.get<std::string>("status") != "ok") {
    std::cerr << "\nERROR: " << pt.get<std::string>("status") << "\n\n";
    throw std::runtime_error("Operation Cancelled");
  }
  int32_t device_id = pt.get<int32_t>("device_info.device_index", -1);
  if (device_id < 0) {
    std::cerr << "\nERROR: Failed to get device index\n\n";
    throw std::runtime_error("Operation Cancelled");
  }
  // RUN TESTCASE
  double fps;
  auto   result = validate::ValidateDevice(device_id, &fps);
  if (result == validate::RESULT::SUCCESS) {
    std::cout << "\n MA35 Validate:  transcode test – PASSED (AVC transcode; " << std::setprecision(4) << fps << " fps)"
              << "\n\n";
  } else {
    pt.put("status", " MA35 Validate:  transcode test – Failed (AVC transcode)");
    std::cerr << "\n MA35 Validate:  transcode test – Failed (AVC transcode)\n\n";
  }

  // -- Write output file ----------------------------------------------
  if (!sOutput.empty()) {
    std::ofstream fOutput;
    fOutput.open(sOutput, std::ios::out | std::ios::binary);
    if (!fOutput.is_open())
      throw std::runtime_error((boost::format("Unable to open the output file '%s' for writing.") % sOutput).str());

    boost::property_tree::json_parser::write_json(fOutput, pt, true /*Pretty Print*/);
    std::cout << boost::format("Successfully wrote the %s file: %s") % sFormat % sOutput << std::endl;
  }
}
