// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#pragma once

#include <unordered_set>
#include <fstream>

#include <nlohmann/json.hpp>
using json   = nlohmann::ordered_json;
using json_t = json;

namespace jsf_int {

/// @brief Base class for jsf nodes
///
/// A jsf service tree is formed by a combination of leaf and internal nodes.  This class is the base class
/// for all nodes in the tree.
class Node {
public:
  using ptr_t = std::unique_ptr<Node>;

public:
  std::string m_name; /// name
  Node*       m_parent = nullptr;
  bool        m_isLeaf;

protected:
  /// construct a node
  explicit Node(const std::string& Name, bool IsLeaf = false) : m_name(Name), m_isLeaf(IsLeaf) {
  }

public:
  virtual ~Node() = default;
};

/// @brief An internal node in the jsf tree
///
/// Any none-leaf node in the service tree is an internal node, which may contain other nodes as children.
class NodeInternal : public Node {
private:
  struct Hasher {
    using is_transparent = void;

    size_t operator()(const Node::ptr_t& Value) const { return std::hash<std::string>()(Value->m_name); }
    size_t operator()(std::string_view Value) const { return std::hash<std::string>()(std::string(Value)); }
  };
  struct Equals {
    using is_transparent = void;
    bool operator()(const Node::ptr_t& Lhs, const Node::ptr_t& Rhs) const { return Lhs->m_name == Rhs->m_name; }
    bool operator()(std::string_view Lhs, const Node::ptr_t& Rhs) const { return Lhs == Rhs->m_name; }
  };
  using children_t = std::unordered_set<Node::ptr_t, Hasher, Equals>;

public:
  children_t m_children;

public:
  explicit NodeInternal(const std::string& Name = {}) : Node(Name) {
  }
};

class NodeLeaf : public Node {
public:
  NodeLeaf(const std::string& Name) : Node(Name, true) {
  }

  virtual json Invoke(const json& Parameters = {}, bool Iteration = false) = 0;

  virtual bool IsPathCapable() const { return false; }
  virtual bool UsesParameters() const { return false; }
  virtual json InvokeWithPath(std::string_view /* Path */, const json& /* Parameters */ = {}) {
    // TODO(jkuhn): better exception
    throw std::logic_error("InvokeWithPath not implemented (but advertised)");
  }
};

class NodeJson : public NodeLeaf {
protected:
  json m_value;

public:
  NodeJson(const std::string& Name, const json& Value) : NodeLeaf(Name), m_value(Value) {
  }

  virtual json Invoke(const json& /* Parameters */ = {}, bool /* Iteration */ = false) override final { // TODO(jkuhn): error on parameters present
    return m_value;
  }
};

class Jsf {
private:
  NodeInternal m_root;

public:
  Jsf() = default;

private:
  Jsf(const Jsf&) = delete;
  Jsf(Jsf&&)      = delete;
  Jsf& operator=(const Jsf&) = delete;
  Jsf& operator=(Jsf&&) = delete;

public:
  Node&                              GetRoot() { return m_root; }
  std::pair<Node*, std::string_view> Find(std::string_view Path, bool Create = false);
  Node*                              Find(const char* Path, bool Create = false) = delete;
  Node*                              Insert(Node* Parent, Node::ptr_t Node);
  Node*                              Insert(std::string_view Path, Node::ptr_t Node);
  Node*                              Insert(const char* Path, Node::ptr_t Node) = delete;
  json                               RecurseBuild(Node* Node, const json& Params);
  std::string                        Execute(const std::string& Request);

  template <typename T, typename... ARGS>
  std::unique_ptr<T> Create(ARGS&&... Args) {
    return std::make_unique<T>(std::forward<ARGS>(Args)...);
  }
};

std::unique_ptr<Jsf> CreateJsf();

} // namespace jsf_int
