// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include "jsf_mautil.h"

#include <sys/mman.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <string_view>
#include <optional>

#include <Logging.h>

#include <jsf_helpers.h>

using namespace nlohmann::json_literals;
using namespace std::literals;

namespace {

std::unique_ptr<jsf_int::Jsf> s_jsf;

// example dynamic node, assume Device will be used to access shared information;

class Device;

class SensorInfoNode : public jsf_helpers::NodeRawFile {
private:
  std::string m_location;
  std::string m_unit;
  unsigned    m_index;
  bool        m_demo;

public:
  SensorInfoNode(const std::string& SysfsRoot, const std::string& Name, const std::string& Location, const std::string& Unit, unsigned Index,
      const std::string& SysfsFile, bool Demo) :
      NodeRawFile(Name, SysfsRoot + "/" + SysfsFile),
      m_location(Location), m_unit(Unit), m_index(Index), m_demo(Demo) {
      }

  virtual json Invoke(const json& Parameters, bool Iteration) override;
};

class ProductInfoNode : public jsf_helpers::NodeRawFile {
private:
  unsigned m_index;
  bool     m_demo;

public:
  ProductInfoNode(const std::string& SysfsRoot, const std::string& Name, unsigned Index, const std::string& SysfsFile, bool Demo) :
      NodeRawFile(Name, SysfsRoot + "/" + SysfsFile), m_index(Index), m_demo(Demo) {
      }

  virtual json Invoke(const json& Parameters, bool Iteration) override;
};

class SkuNumber : public ProductInfoNode {
private:
  bool m_demo;

public:
  SkuNumber(const std::string& SysfsRoot, const std::string& Name, unsigned Index, const std::string& SysfsFile, bool Demo) :
      ProductInfoNode(SysfsRoot, Name, Index, SysfsFile, Demo), m_demo(Demo) {
      }
  virtual json Invoke(const json& Parameters, bool Iteration) override;
};

class VersionInfoNode : public ProductInfoNode {
private:
  bool m_demo;

public:
  VersionInfoNode(const std::string& SysfsRoot, const std::string& Name, unsigned Index, const std::string& SysfsFile, bool Demo) :
      ProductInfoNode(SysfsRoot, Name, Index, SysfsFile, Demo), m_demo(Demo) {
      }

  virtual json Invoke(const json& Parameters, bool Iteration) override;
};

class VariableValNode : public jsf_helpers::NodeRawFile {
private:
  bool m_demo;

public:
  VariableValNode(const std::string& SysfsRoot, const std::string& Name, const std::string& SysfsFile, bool Demo) :
      NodeRawFile(Name, SysfsRoot + "/" + SysfsFile), m_demo(Demo) {
      }
  virtual json Invoke(const json& Parameters, bool Iteration) override;
};

class NameValPairNode : public jsf_helpers::NodeRawFile {
private:
  unsigned m_index;
  bool     m_demo;

public:
  NameValPairNode(const std::string& SysfsRoot, const std::string& Name, unsigned Index, const std::string& SysfsFile, bool Demo) :
      NodeRawFile(Name, SysfsRoot + "/" + SysfsFile), m_index(Index), m_demo(Demo) {
      }

  virtual json Invoke(const json& Parameters, bool Iteration) override;
};

class ErrorCounters : public jsf_helpers::NodeRawFile {
private:
  unsigned m_index;
  bool     m_demo;

public:
  ErrorCounters(const std::string& SysfsRoot, const std::string& Name, unsigned Index, const std::string& SysfsFile, bool Demo) :
      NodeRawFile(Name, SysfsRoot + "/" + SysfsFile), m_index(Index), m_demo(Demo) {
      }

  virtual json Invoke(const json& Parameters, bool Iteration) override;
};

class DeviceMemory : public jsf_helpers::NodeRawFile {
private:
  unsigned m_index;
  bool     m_demo;

public:
  DeviceMemory(const std::string& SysfsRoot, const std::string& Name, unsigned Index, const std::string& SysfsFile, bool Demo) :
      NodeRawFile(Name, SysfsRoot + "/" + SysfsFile), m_index(Index), m_demo(Demo) {
      }

  virtual json Invoke(const json& Parameters, bool Iteration) override;
};

class DdrInfoNode : public jsf_helpers::NodeRawFile {
private:
  bool m_demo;

public:
  DdrInfoNode(const std::string& SysfsRoot, const std::string& Name, const std::string& SysfsFile, bool Demo) :
      NodeRawFile(Name, SysfsRoot + "/" + SysfsFile), m_demo(Demo) {
      }

  virtual json Invoke(const json& Parameters, bool Iteration) override;
};

class Reset : public jsf_int::NodeLeaf {
private:
  Device& m_device;
  bool    m_demo;

public:
  Reset(Device& Device, const std::string& Name, bool Demo) : NodeLeaf(Name), m_device(Device), m_demo(Demo) {
  }

  virtual json Invoke(const json& Parameters = {}, bool Iteration = false) override;
};

class Device : public jsf_int::NodeInternal {
private:
  pci_dev     m_pciDevice;
  std::string m_dbdf;
  unsigned    m_deviceId;
  std::string m_sysfsRoot;

public:
  explicit Device(const pci_dev& PciDevice, unsigned DeviceNum, bool Demo = false) :
      NodeInternal(jsf_helpers::DbdfString(PciDevice)), m_pciDevice(PciDevice), m_dbdf(jsf_helpers::DbdfString(PciDevice)), m_deviceId(PciDevice.device_id),
      m_sysfsRoot("/sys/class/misc/ama_transcoder"s + std::to_string(DeviceNum)) {
    s_jsf->Insert(this, s_jsf->Create<Reset>(*this, "reset", Demo)); // Reset (testapps/jmautil --live devices/BDF/reset '"type":flr')

    auto devNode = s_jsf->Insert(this, s_jsf->Create<jsf_int::NodeInternal>("device_info"s)); // MA35 Device Info

    s_jsf->Insert(devNode, s_jsf->Create<jsf_int::NodeJson>("device_index", DeviceNum));
    s_jsf->Insert(devNode, s_jsf->Create<jsf_int::NodeJson>("processor_type", "VPU (Video Processing Unit)"));
    s_jsf->Insert(devNode, s_jsf->Create<ProductInfoNode>(m_sysfsRoot, "product_name"s, 0, "device_info"s, Demo));
    s_jsf->Insert(devNode, s_jsf->Create<ProductInfoNode>(m_sysfsRoot, "product_revision"s, 1, "device_info"s, Demo));
    s_jsf->Insert(devNode, s_jsf->Create<ProductInfoNode>(m_sysfsRoot, "product_serial_number"s, 2, "device_info"s, Demo));
    // s_jsf->Insert(devNode, s_jsf->Create<ProductInfoNode>(*this, "max_power_mode"s, 9, "device_info"s, Demo));

    auto chipIdNode = s_jsf->Insert(devNode, s_jsf->Create<jsf_int::NodeInternal>("chip_id"s)); // MA35 HW Id's
    s_jsf->Insert(chipIdNode, s_jsf->Create<ProductInfoNode>(m_sysfsRoot, "part_number"s, 3, "device_info"s, Demo));
    s_jsf->Insert(chipIdNode, s_jsf->Create<SkuNumber>(m_sysfsRoot, "sku_number"s, 3, "device_info"s, Demo)); // Last two digit of part_number
    s_jsf->Insert(chipIdNode, s_jsf->Create<ProductInfoNode>(m_sysfsRoot, "oem_id"s, 4, "device_info"s, Demo));
    // s_jsf->Insert(hwInfoNode, s_jsf->Create<jsf_int::NodeJson>("chip_type", ""));

    auto thresholdNode = s_jsf->Insert(devNode, s_jsf->Create<jsf_int::NodeInternal>("thresholds"s)); // MA35 Thresholds
    s_jsf->Insert(thresholdNode, s_jsf->Create<jsf_int::NodeJson>("threshold_temp_C", 85));           // BMC ASIC Junction
    s_jsf->Insert(thresholdNode, s_jsf->Create<jsf_int::NodeJson>("max_operating_temp_C", 105));      // Warning Limit ASIC Junction
    s_jsf->Insert(thresholdNode, s_jsf->Create<jsf_int::NodeJson>("shutdown_temp_C", 110));           // Fatal Limit ASIC Junction

    auto verNode = s_jsf->Insert(devNode, s_jsf->Create<jsf_int::NodeInternal>("version_info"s)); // MA35 Version Info
    s_jsf->Insert(verNode, s_jsf->Create<VersionInfoNode>(m_sysfsRoot, "ZSP"s, 0, "version_information"s, Demo));
    s_jsf->Insert(verNode, s_jsf->Create<VersionInfoNode>(m_sysfsRoot, "SC"s, 1, "version_information"s, Demo));
    s_jsf->Insert(verNode, s_jsf->Create<VersionInfoNode>(m_sysfsRoot, "eSecure"s, 2, "version_information"s, Demo));
    s_jsf->Insert(verNode, s_jsf->Create<VersionInfoNode>(m_sysfsRoot, "PCIe"s, 3, "version_information"s, Demo));
    s_jsf->Insert(verNode, s_jsf->Create<VersionInfoNode>(m_sysfsRoot, "PcieCtlPatch"s, 4, "version_information"s, Demo));
    s_jsf->Insert(verNode, s_jsf->Create<VersionInfoNode>(m_sysfsRoot, "PciePhyPatch"s, 5, "version_information"s, Demo));

    auto pcieNode = s_jsf->Insert(this, s_jsf->Create<jsf_int::NodeInternal>("pcie_info"s)); // Pcie Info
    s_jsf->Insert(pcieNode, s_jsf->Create<ProductInfoNode>(m_sysfsRoot, "vendor_id"s, 5, "device_info"s, Demo));
    s_jsf->Insert(pcieNode, s_jsf->Create<ProductInfoNode>(m_sysfsRoot, "device_id"s, 6, "device_info"s, Demo));
    s_jsf->Insert(pcieNode, s_jsf->Create<ProductInfoNode>(m_sysfsRoot, "speed"s, 0, "link_status"s, Demo));
    s_jsf->Insert(pcieNode, s_jsf->Create<ProductInfoNode>(m_sysfsRoot, "width"s, 1, "link_status"s, Demo));
    s_jsf->Insert(pcieNode, s_jsf->Create<VariableValNode>(m_sysfsRoot, "tx_bw_MBps"s, "pcie_rc2ep_bw"s, Demo));
    s_jsf->Insert(pcieNode, s_jsf->Create<VariableValNode>(m_sysfsRoot, "rx_bw_MBps"s, "pcie_ep2rc_bw"s, Demo));

    s_jsf->Insert(devNode, s_jsf->Create<VariableValNode>(m_sysfsRoot, "uptime_S"s, "uptime_s"s, Demo));

    auto telNode = s_jsf->Insert(this, s_jsf->Create<jsf_int::NodeInternal>("telemetry"s));

    auto deviceMemNode = s_jsf->Insert(telNode, s_jsf->Create<jsf_int::NodeInternal>("device_memory"s)); // Device Memory
#if defined(OSAL)
    s_jsf->Insert(deviceMemNode, s_jsf->Create<DeviceMemory>(m_sysfsRoot, "slice0"s, 0, "device_mem_info"s, Demo));
    s_jsf->Insert(deviceMemNode, s_jsf->Create<DeviceMemory>(m_sysfsRoot, "slice1"s, 1, "device_mem_info"s, Demo));
    s_jsf->Insert(deviceMemNode, s_jsf->Create<DeviceMemory>(m_sysfsRoot, "mmio"s, 2, "device_mem_info"s, Demo));
#else
    s_jsf->Insert(deviceMemNode, s_jsf->Create<DeviceMemory>(m_sysfsRoot, "slice0"s, 0, "mem_info"s, Demo));
    s_jsf->Insert(deviceMemNode, s_jsf->Create<DeviceMemory>(m_sysfsRoot, "slice1"s, 1, "mem_info"s, Demo));
#endif

    auto ddrBwNode = s_jsf->Insert(telNode, s_jsf->Create<jsf_int::NodeInternal>("ddr_bandwidth"s)); // DDR Bandwidth
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(m_sysfsRoot, "s1_axi_r_MBps"s, "ddrbw_s1_axi_r_MBps"s, Demo));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(m_sysfsRoot, "s1_axi_w_MBps"s, "ddrbw_s1_axi_w_MBps"s, Demo));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(m_sysfsRoot, "s1_dfi_r_MBps"s, "ddrbw_s1_dfi_r_MBps"s, Demo));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(m_sysfsRoot, "s1_dfi_w_MBps"s, "ddrbw_s1_dfi_w_MBps"s, Demo));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(m_sysfsRoot, "s2_axi_r_MBps"s, "ddrbw_s2_axi_r_MBps"s, Demo));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(m_sysfsRoot, "s2_axi_w_MBps"s, "ddrbw_s2_axi_w_MBps"s, Demo));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(m_sysfsRoot, "s2_dfi_r_MBps"s, "ddrbw_s2_dfi_r_MBps"s, Demo));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(m_sysfsRoot, "s2_dfi_w_MBps"s, "ddrbw_s2_dfi_w_MBps"s, Demo));

    auto errCounterNode = s_jsf->Insert(telNode, s_jsf->Create<jsf_int::NodeInternal>("error_counters")); // Error Counters
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(m_sysfsRoot, "pcie", 0, "pcie_error_information", Demo));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(m_sysfsRoot, "ddr_ch0", 0, "ddr_error_information", Demo));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(m_sysfsRoot, "ddr_ch1", 2, "ddr_error_information", Demo));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(m_sysfsRoot, "ddr_ch2", 4, "ddr_error_information", Demo));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(m_sysfsRoot, "ddr_ch3", 6, "ddr_error_information", Demo));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(m_sysfsRoot, "ddr_ch4", 8, "ddr_error_information", Demo));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(m_sysfsRoot, "ddr_ch5", 10, "ddr_error_information", Demo));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(m_sysfsRoot, "ddr_ch6", 12, "ddr_error_information", Demo));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(m_sysfsRoot, "ddr_ch7", 14, "ddr_error_information", Demo));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(m_sysfsRoot, "THS1_axi_sram", 0, "axi_sram_error_information", Demo));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(m_sysfsRoot, "THS2_axi_sram", 2, "axi_sram_error_information", Demo));

    // Chip level Telemetry
    auto asicNode = s_jsf->Insert(telNode, s_jsf->Create<jsf_int::NodeInternal>("chip"s)); // sn_chip

    auto voltNode = s_jsf->Insert(asicNode, s_jsf->Create<jsf_int::NodeInternal>("pvt_voltage")); // Voltage
    s_jsf->Insert(voltNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor0"s, "enc"s, "volt_mV"s, 0, "pvt_voltage"s, Demo));
    s_jsf->Insert(voltNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor1"s, "ml_engine"s, "volt_mV"s, 1, "pvt_voltage"s, Demo));
    s_jsf->Insert(voltNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor2"s, "ddr0"s, "volt_mV"s, 2, "pvt_voltage"s, Demo));
    s_jsf->Insert(voltNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor3"s, "aux"s, "volt_mV"s, 3, "pvt_voltage"s, Demo));

    auto tempNode = s_jsf->Insert(asicNode, s_jsf->Create<jsf_int::NodeInternal>("pvt_temperature")); // Temperature
    s_jsf->Insert(tempNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor0"s, "ma35_temp_s0"s, "temp_C"s, 0, "pvt_temperature"s, Demo));
    s_jsf->Insert(tempNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor1"s, "ma35_temp_s1"s, "temp_C"s, 1, "pvt_temperature"s, Demo));
    s_jsf->Insert(tempNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor2"s, "ma35_temp_s2"s, "temp_C"s, 2, "pvt_temperature"s, Demo));
    s_jsf->Insert(tempNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor3"s, "ma35_temp_s3"s, "temp_C"s, 3, "pvt_temperature"s, Demo));

    // Board level Telemetry
    auto boardNode = s_jsf->Insert(telNode, s_jsf->Create<jsf_int::NodeInternal>("board"s));

    auto bTemprNode = s_jsf->Insert(boardNode, s_jsf->Create<jsf_int::NodeInternal>("temperature")); // Board Temperature
    s_jsf->Insert(bTemprNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "board_temp"s, "board_temp"s, "temp_C", 0, "board_temp"s, Demo));

    auto bPowerNode = s_jsf->Insert(boardNode, s_jsf->Create<jsf_int::NodeInternal>("power")); // Board Temperature
    s_jsf->Insert(bPowerNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "board_power"s, "board_power"s, "power_mW", 0, "board_power"s, Demo));

    auto bVoltNode = s_jsf->Insert(boardNode, s_jsf->Create<jsf_int::NodeInternal>("voltage_current")); // Board Voltage
    s_jsf->Insert(bVoltNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor5"s, "3V PEX Voltage"s, "volt_mV", 0, "board_vc"s, Demo));
    s_jsf->Insert(bVoltNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor4"s, "3V AUX Voltage"s, "volt_mV", 1, "board_vc"s, Demo));
    s_jsf->Insert(bVoltNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor3"s, "12V PEX Voltage"s, "volt_mV", 2, "board_vc"s, Demo));
    s_jsf->Insert(bVoltNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor2"s, "3V PEX Current"s, "current_mA", 3, "board_vc"s, Demo));
    s_jsf->Insert(bVoltNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor1"s, "3V AUX Current"s, "current_mA", 4, "board_vc"s, Demo));
    s_jsf->Insert(bVoltNode, s_jsf->Create<SensorInfoNode>(m_sysfsRoot, "sensor0"s, "12V PEX Current"s, "current_mA", 5, "board_vc"s, Demo));
  }

  void         fail() { abort(); }
  std::string  GetBdf() const { return m_dbdf; }
  unsigned int GetDeviceId() const { return m_deviceId; }
};

json Reset::Invoke(const json& Parameters, bool Iteration) {
  if (Iteration) {
    return {};
  }

  const auto bdf       = m_device.GetBdf();
  unsigned   device_id = m_device.GetDeviceId();

  if (device_id != 0x5071) { // VF device_id: 0x5071
    throw std::runtime_error("Invalid VF device, please try \"mautil reset\" with a valid VF device from \"Devices present\"");
  }

  auto bdf_pf        = bdf;
  bdf_pf             = bdf_pf.substr(0, bdf_pf.size() - 1) + "0";
  auto sriov_pf_file = "/sys/bus/pci/devices/" + bdf_pf + "/sriov_numvfs";
  if (std::filesystem::exists(sriov_pf_file) == 1) { // Check if mautil reset is triggered from Host (mautil should be used for VF device reset under VM)
    throw std::runtime_error("\"mautil\" can be used to reset VF device under VM (use \"mamgmt\" to reset under host)");
  }

  if (Parameters.contains("type")) { // flr
    if (Parameters["type"] == "flr") {

      const auto sysfs_unbind = "/sys/bus/pci/drivers/transcoder/unbind"; // echo 0000:01:00.1 > /sys/bus/pci/drivers/transcoder/unbind
      auto       fd_unbind    = open(sysfs_unbind, O_WRONLY);
      if (fd_unbind >= 0) {
        auto ret = write(fd_unbind, bdf.c_str(), sizeof(bdf));
        if (ret != sizeof(bdf)) {
          logFatal("unbind: write failed");
          close(fd_unbind);
          m_device.fail();
        }
        close(fd_unbind);
      }

      const auto    sysfs_reset = "/sys/bus/pci/devices/" + bdf + "/reset"; // echo 1 > /sys/bus/pci/devices/0000\:01\:00.1/reset
      unsigned char val_reset   = 1;
      auto          fd_reset    = open(sysfs_reset.c_str(), O_WRONLY);
      if (fd_reset < 0) {
        logFatal("reset: file Not opened");
        m_device.fail();
      }
      auto ret = write(fd_reset, std::to_string(val_reset).c_str(), 1);
      if (ret != 1) {
        logFatal("reset: write failed");
        close(fd_reset);
        m_device.fail();
      }
      close(fd_reset);

      const auto sysfs_bind = "/sys/bus/pci/drivers/transcoder/bind"; // echo 0000:01:00.1 > /sys/bus/pci/drivers/transcoder/bind
      auto       fd_bind    = open(sysfs_bind, O_WRONLY);
      if (fd_bind >= 0) {
        ret = write(fd_bind, bdf.c_str(), sizeof(bdf));
        if (ret != sizeof(bdf)) {
          logFatal("bind: write failed");
          close(fd_bind);
          m_device.fail();
        }
        close(fd_bind);
      }

      return 1;
    }
    throw std::runtime_error("invalid reset type");
  }

  if (!Parameters.is_null()) {
    throw std::runtime_error("reset type required / invalid type");
  }
  return {};
}

template <typename T = uint32_t>
requires(std::is_integral_v<T>) std::optional<T> GetVal(const std::optional<std::string>& Val, unsigned Count = 0) {
  if (!Val) {
    return {};
  }
  unsigned pos = 0;
  for (unsigned i = 0; i < Count; ++i) {
    auto eq = Val->find_first_of('=', pos);
    if (eq == std::string::npos) {
      return {};
    }
    pos = eq + 1;
  }
  try {
    return T(std::stoll(Val->substr(pos)));
  } catch (...) {
    return {};
  }
}

json NameValPairNode::Invoke(const json& Parameters, bool Iteration) {
  if (m_demo) {
    return 400 + (rand() & 255);
  }
  auto val = GetVal(NodeRawFile::Invoke(Parameters, Iteration), m_index + 1);
  if (!val) {
    return "ERROR"s;
  }
  return *val;
}

json ErrorCounters::Invoke(const json& Parameters, bool Iteration) {
  json err_counters;
  if (m_demo) {
    err_counters["uncorrectable"] = 400 + (rand() & 255);
    err_counters["correctable"]   = 400 + (rand() & 255);
    return err_counters;
  }

  for (unsigned int count = m_index; count < (m_index + 2);) {
    auto val = GetVal(NodeRawFile::Invoke(Parameters, Iteration), ++count);
    if (!val) {
      if (count % 2) {
        err_counters["uncorrectable"] = "ERROR";
      } else {
        err_counters["correctable"] = "ERROR";
      }
    } else {
      if (count % 2) {
        err_counters["uncorrectable"] = *val;
      } else {
        err_counters["correctable"] = *val;
      }
    }
  }

  return err_counters;
}

json SensorInfoNode::Invoke(const json& Parameters, bool Iteration) {
  json sensor_info;
  if (m_demo) {
    sensor_info["id"]   = m_location;
    sensor_info[m_unit] = 60 + (rand() & 71) - 16;
    return sensor_info;
  }

  auto val = GetVal(NodeRawFile::Invoke(Parameters, Iteration), m_index + 1);
  if (!val) {
    return "ERROR"s;
  }
  sensor_info["id"]   = m_location;
  sensor_info[m_unit] = *val;
  return sensor_info;
}

json ProductInfoNode::Invoke(const json& Parameters, bool Iteration) {
  if (m_demo) {
    return std::to_string(400 + (rand() & 255));
  }

  std::string val = NodeRawFile::Invoke(Parameters, Iteration);
  if (!val.empty()) {
    unsigned pos = 0;
    for (unsigned i = 0; i < (m_index + 1); ++i) {
      auto eq = val.find_first_of('=', pos);
      if (eq == std::string::npos) {
        return {};
      }
      pos = eq + 2;
    }
    val = val.substr(pos, val.find_first_of('\n', pos) - pos);
    if (auto sub_pos = val.find_first_of(',', 0); sub_pos != std::string::npos) {
      val.resize(sub_pos);
    }
  }
  return (val.empty()) ? "ERROR"s : val;
}

json SkuNumber::Invoke(const json& Parameters, bool Iteration) {
  if (m_demo) {
    return std::to_string(400 + (rand() & 255));
  }
  std::string part_no = ProductInfoNode::Invoke(Parameters, Iteration);
  auto        sku_pos = part_no.find_first_of('-') + 1;
  if (sku_pos == std::string::npos) {
    return {};
  }
  auto sku_no = part_no.substr(sku_pos, part_no.find_first_of('\n', sku_pos) - sku_pos);
  return (sku_no.empty()) ? "ERROR"s : sku_no;
}

json VersionInfoNode::Invoke(const json& Parameters, bool Iteration) {
  if (m_demo) {
    return std::to_string(rand() & 9) + "." + std::to_string(rand() & 9) + "." + std::to_string(rand() & 9);
  }
  return ProductInfoNode::Invoke(Parameters, Iteration);
}

json DeviceMemory::Invoke(const json& Parameters, bool Iteration) { // region , base_addr, used, size , %used
  json device_mem_info = {};
  if (m_demo) {
    device_mem_info["size_MB"] = 600ull + (rand() & 255);
    device_mem_info["used_MB"] = 100ull + (rand() & 255);
    device_mem_info["used%"]   = 1 + rand() % 100;
    return device_mem_info;
  }

  std::string val = NodeRawFile::Invoke(Parameters, Iteration);
  if (!val.empty()) {
    std::stringstream val_stream(val.c_str());
    std::string       line;

    for (unsigned index = 0; index < (m_index + 1); ++index) {
      std::getline(val_stream, line);
    }
    std::istringstream       line_stream(line);
    std::vector<std::string> tokens(std::istream_iterator<std::string>{line_stream}, std::istream_iterator<std::string>{});

#if defined(USE_MEMORY_OSAL)
    auto used                  = std::stoull(tokens[2].c_str()) * 4096 / 1024 / 1024; // device_mem_info
    auto size                  = std::stoull(tokens[3].c_str()) * 4096 / 1024 / 1024; // device_mem_info
    auto used_percent          = (used * 100.0) / size;
    device_mem_info["size_MB"] = size;
    device_mem_info["used_MB"] = used;
    device_mem_info["used%"]   = used_percent;
#else
    auto used         = std::stoul(tokens[1].c_str());
    auto free         = std::stoul(tokens[4].c_str());
    auto used_percent = std::stof(tokens[7].c_str());
    auto size         = used + free;
#endif
    device_mem_info["size_MB"] = size;
    device_mem_info["used_MB"] = used;
    device_mem_info["used%"]   = used_percent;
  }
  return device_mem_info;
}

json DdrInfoNode::Invoke(const json& Parameters, bool Iteration) {
  json ddr_bw = {};
  if (m_demo) {
    ddr_bw["instantaneous"] = 200 + (rand() & 1023);
    ddr_bw["peak"]          = 300 + (rand() & 1023);
    return ddr_bw;
  }
  std::string val       = NodeRawFile::Invoke(Parameters, Iteration);
  unsigned    ddrBwInst = 0;
  // unsigned    ddrBwPeak = 0;
  if (!val.empty()) {
    std::stringstream stream(val.c_str());
    int               ddr_bw_per_channel;
    while (stream >> ddr_bw_per_channel) {
      ddrBwInst += ddr_bw_per_channel;
    }
  }
  ddr_bw["instantaneous"] = ddrBwInst;
  // ddr_bw["peak"]          = ddrBwPeak;

  return ddr_bw;
}

json VariableValNode::Invoke(const json& Parameters, bool Iteration) {
  if (m_demo) {
    return rand();
  }
  auto val = GetVal(NodeRawFile::Invoke(Parameters, Iteration));
  if (!val) {
    return "ERROR"s;
  }
  return *val;
}

void DemoInit(jsf_int::Node* Node) {
  s_jsf->Insert("version"sv, s_jsf->Create<jsf_int::NodeJson>("pkg_version", "0.0.1"));

  srand(time(nullptr));
  pci_dev dummy   = {};
  dummy.vendor_id = 0x10ee;
  dummy.device_id = 0x5070;
  dummy.bus       = 4;
  s_jsf->Insert(Node, s_jsf->Create<Device>(dummy, 0, true));
  dummy.bus = 5;
  s_jsf->Insert(Node, s_jsf->Create<Device>(dummy, 1, true));

  static const std::vector<const char*> sDevices = {"0000:04:00.0", "0000:05:00.0"};
  s_jsf->Insert(""sv, s_jsf->Create<jsf_int::NodeJson>("device_list", sDevices));
}

} // namespace

#if 0
void TestReset() {
  int vf_device = open("/dev/ama_transcoder_vf0", O_RDWR); // Initialize m_device
  if (vf_device < 0) {
    logFatal("device open failed: /dev/ama_transcoder_vf0");
    abort();
  }
  auto vf_vaddr = mmap(NULL, 0x8000000, PROT_READ | PROT_WRITE, MAP_SHARED, vf_device, 0xec000000); // VF (VM)

  if (vf_vaddr == MAP_FAILED) {
    logFatal("mmap failed for VF\n");
    abort();
  }

  // XABR-DFT Write
  volatile uint32_t* vf_paddr       = reinterpret_cast<volatile uint32_t*>(reinterpret_cast<volatile uint8_t*>(vf_vaddr) + 0x1c0000 + 0xa0); // DFT Pins
  uint32_t           vf_XabrDftPins = *vf_paddr;
  logNotice("VF_XabrDftPins: 0x%08x", vf_XabrDftPins);

  uint32_t xval                                                                                              = 0x6;
  *(reinterpret_cast<volatile uint32_t*>((reinterpret_cast<volatile uint8_t*>(vf_vaddr) + 0x1c0000 + 0xa0))) = xval;
  // uint32_t vf_XabrDftPins2                                                                                   = *vf_paddr;
  // logNotice("VF_XabrDftPins (after) : 0x%08x", vf_XabrDftPins2);

  // vf_paddr   = reinterpret_cast<volatile uint32_t*>(reinterpret_cast<volatile uint8_t*>(vf_vaddr) + 0x1c0000 + 0xb0); // XABR Version
  // uint32_t           vf_XabrVer = *vf_paddr;
  // logNotice("VF_XabrVer: 0x%08x", vf_XabrVer);
}
#endif

namespace jsf_mautil {

void init(bool Demo) {
  s_jsf = jsf_int::CreateJsf();

  auto devicesNode = s_jsf->Insert(""sv, s_jsf->Create<jsf_int::NodeInternal>("devices"));
  s_jsf->Insert(""sv, s_jsf->Create<jsf_helpers::NodeSystemInfo>("host"));

  if (Demo) {
    DemoInit(devicesNode);
    return;
  }

  s_jsf->Insert("version"sv, s_jsf->Create<jsf_helpers::NodeVersionInfo>("pkg_version", "/opt/amd/ama/ma35/version"));

  std::vector<std::string> deviceList;
  for (const auto deviceId : {0x5070, 0x5071}) {
    auto devices = jsf_helpers::GetPciDevices(0x10ee, deviceId);
    for (const auto& device : devices) {
      int  device_num = -1;
      auto bdf        = jsf_helpers::DbdfString(device);

      if (deviceId == 0x5070) {
        const auto    sriov_pf_path = "/sys/bus/pci/devices/" + bdf + "/sriov_numvfs";
        std::ifstream sriov_pf_file(sriov_pf_path);
        int           val_sriov_numvfs = 0;
        sriov_pf_file >> val_sriov_numvfs;
        if (val_sriov_numvfs) {
          continue;
        }
      }

      for (unsigned dev_cnt = 0; dev_cnt < jsf_helpers::MAX_TRANS_DEVICES; ++dev_cnt) {
        auto          path = "/sys/class/misc/ama_transcoder"s + std::to_string(dev_cnt) + "/bus_id";
        std::ifstream stream(path);
        if (!stream.is_open()) {
          continue;
        }
        std::string transBdf;
        std::getline(stream, transBdf);
        if (bdf != transBdf) {
          continue;
        }
        device_num = dev_cnt;
        break;
      }
      if (device_num == -1) {
        logNotice("Transcoder device does not exist for BDF: %s", bdf.c_str());
        continue;
      }
      s_jsf->Insert(devicesNode, s_jsf->Create<Device>(device, device_num));
      deviceList.push_back(bdf);
    }
  }
  s_jsf->Insert(""sv, s_jsf->Create<jsf_int::NodeJson>("device_list", deviceList));
}

void destroy() {
  s_jsf.reset();
}

std::string execute(const std::string& Request) {
  return s_jsf->Execute(Request);
}

} // namespace jsf_mautil
